# -*- coding: utf-8 -*-
import functools

from jsonschema import validators
from jsonschema.exceptions import ValidationError
from jsonschema.validators import Draft4Validator
from swagger_spec_validator.ref_validators import in_scope

from bravado_core.model import MODEL_MARKER
from bravado_core.schema import is_param_spec
from bravado_core.schema import is_prop_nullable
from bravado_core.schema import is_required
from bravado_core.util import memoize_by_id


"""Draft4Validator is not completely compatible with Swagger 2.0 schema
objects like parameter, etc. Swagger20Validator is an extension of
Draft4Validator which customizes/wraps some of the operations of the default
validator.
"""

# Extract "original" validators as defined in jsonschema library. Saving those to constants to reduce indirections during validation.  # noqa
_DRAFT4_ENUM_VALIDATOR = Draft4Validator.VALIDATORS['enum']
_DRAFT4_FORMAT_VALIDATOR = Draft4Validator.VALIDATORS['format']
_DRAFT4_REQUIRED_VALIDATOR = Draft4Validator.VALIDATORS['required']
_DRAFT4_TYPE_VALIDATOR = Draft4Validator.VALIDATORS['type']


def format_validator(swagger_spec, validator, format, instance, schema):
    """Skip the `format` validator when a Swagger parameter value is None.
    Otherwise it will fail with a "Failed validating u'format' in schema" failure instead
    of letting the downstream `required_validator` do its job.
    Also skip when a Swagger property value is None and the schema contains
    the extension field `x-nullable` set to True.
    In all other cases, delegate to the existing Draft4 `format` validator.

    """
    if instance is None and (
        is_param_spec(swagger_spec, schema) or
        is_prop_nullable(swagger_spec, schema)
    ):
        return

    for error in _DRAFT4_FORMAT_VALIDATOR(validator, format, instance, schema):
        yield error


def type_validator(swagger_spec, validator, types, instance, schema):
    """Skip the `type` validator when a Swagger parameter value is None.
    Otherwise it will fail with a "None is not a valid type" failure instead
    of letting the downstream `required_validator` do its job.
    Also skip when a Swagger property value is None and the schema contains
    the extension field `x-nullable` set to True.
    In all other cases, delegate to the existing Draft4 `type` validator.

    :param swagger_spec: needed for access to deref()
    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :param validator: Validator class used to validate the object
    :type validator: :class:`Swagger20Validator` or
        :class:`jsonschema.validators.Draft4Validator`
    :param types: validate types
    :type types: string or list
    :param instance: object instance value
    :param schema: swagger spec for the object
    :type schema: dict
    """
    if instance is None and (
        is_param_spec(swagger_spec, schema) or
        is_prop_nullable(swagger_spec, schema)
    ):
        return

    for error in _DRAFT4_TYPE_VALIDATOR(validator, types, instance, schema):
        yield error


def required_validator(swagger_spec, validator, required, instance, schema):
    """Swagger 2.0 expects `required` to be a bool in the Parameter object,
    but a list of properties everywhere else.

    :param swagger_spec: needed for access to deref()
    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :param validator: Validator class used to validate the object
    :type validator: :class:`Swagger20Validator` or
        :class:`jsonschema.validators.Draft4Validator`
    :param required: value of `required` field
    :type required: boolean or list or None
    :param instance: object instance value
    :param schema: swagger spec for the object
    :type schema: dict
    """
    if is_param_spec(swagger_spec, schema):
        if required and instance is None:
            yield ValidationError('{0} is a required parameter.'.format(schema['name']))
    else:
        for error in _DRAFT4_REQUIRED_VALIDATOR(validator, required, instance, schema):
            yield error


def enum_validator(swagger_spec, validator, enums, instance, schema):
    """Swagger 2.0 allows enums to be validated against objects of type
    arrays, like query parameter (collectionFormat: multi)

    :param swagger_spec: needed for access to deref()
    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :param validator: Validator class used to validate the object
    :type validator: :class: `Swagger20Validator` or
        :class: `jsonschema.validators.Draft4Validator`
    :param enums: allowed enum values
    :type enums: list
    :param instance: enum instance value
    :param schema: swagger spec for the object
    :type schema: dict
    """

    if instance is None and is_prop_nullable(swagger_spec, schema):
        return

    if schema.get('type') == 'array':
        for element in instance:
            for error in _DRAFT4_ENUM_VALIDATOR(validator, enums, element, schema):
                yield error
        return

    # Handle optional enum params with no value
    if is_param_spec(swagger_spec, schema):
        if instance is None and not is_required(swagger_spec, schema):
            return

    for error in _DRAFT4_ENUM_VALIDATOR(validator, enums, instance, schema):
        yield error


def discriminator_validator(swagger_spec, validator, discriminator_attribute, instance, schema):
    """
    Validates instance against the schema defined by the discriminator attribute.

    [Swagger 2.0 Schema Object](http://swagger.io/specification/#schemaObject) allows discriminator field to be defined.
    discriminator field defines the attribute that will be used to discriminate the object type.

    NOTE: discriminator_validator assumes that discriminator_attribute is not None or empty

    :param swagger_spec: needed for access to deref()
    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :param validator: Validator class used to validate the object
    :type validator: :class: `Swagger20Validator` or
        :class: `jsonschema.validators.Draft4Validator`
    :param discriminator_attribute: name of the discriminator attribute
    :type discriminator_attribute: str
    :param instance: object instance value
    :type instance: dict
    :param schema: swagger spec for the object
    :type schema: dict
    """

    try:
        discriminator_value = instance[discriminator_attribute]
    except KeyError:
        raise ValidationError("'{}' is a required property".format(discriminator_attribute))

    if discriminator_value not in swagger_spec.definitions:
        raise ValidationError(
            message='\'{}\' is not a recognized schema'.format(discriminator_value),
        )

    if discriminator_value == schema[MODEL_MARKER]:
        return

    discriminated_schema = swagger_spec.definitions[discriminator_value]._model_spec
    if 'allOf' not in discriminated_schema:
        raise ValidationError(
            message='discriminated schema \'{}\' must inherit from \'{}\''.format(
                discriminator_value, schema[MODEL_MARKER],
            ),
        )

    schemas_to_remove = [s for s in discriminated_schema['allOf'] if swagger_spec.deref(s) == schema]
    if not schemas_to_remove:
        # Not checking against len(schemas_to_remove) > 1 because it should be prevented by swagger spec validation
        raise ValidationError(
            message='discriminated schema \'{}\' must inherit from \'{}\''.format(
                discriminator_value, schema[MODEL_MARKER],
            ),
        )

    # Remove the current schema from the allOf list in order to avoid unbounded recursion
    # (the current object is already validated against schema)
    # WARNING: This is especially important if internally_dereference_refs is set to true
    #   as we're modifying new_schema and new_schema is a dict (so mutable) we need to copy
    #   it in order to have a brand new dictionary that we can modify
    new_schema = discriminated_schema.copy()
    new_schema['allOf'] = [
        all_of_schema if all_of_schema not in schemas_to_remove else {}
        for all_of_schema in new_schema['allOf']
    ]

    from bravado_core.validate import validate_object  # Local import due to circular dependency
    validate_object(swagger_spec=swagger_spec, object_spec=new_schema, value=instance)


def ref_validator(validator, ref, instance, schema):
    """When validating a request or response that contain $refs, jsonschema's
     RefResolver only contains scope (RefResolver._scopes_stack) for the
     request_spec or response_spec that it is fed. This scope doesn't contain
     the full scope built when ingesting the spec from its root
     (#/ in swagger.json). So, we need to modify the behavior of ref
     validation to use the `x-scope` annotations that were created during spec
     ingestion (see model_discovery in bravado_core/model.py).

    :param validator: Validator class used to validate the object
    :type validator: :class: `Swagger20Validator` or
        :class: `jsonschema.validators.Draft4Validator`
    :param ref: the target of the ref. eg. #/foo/bar/Baz
    :type ref: string
    :param instance: the object being validated
    :param schema: swagger spec that contains the ref.
        eg {'$ref': '#/foo/bar/Baz'}
    :type schema: dict
    """
    # This is a copy of jsonschema._validators.ref(..) with the
    # in_scope(..) context manager applied before any refs are resolved.
    resolve = getattr(validator.resolver, "resolve")
    if resolve is None:
        with in_scope(validator.resolver, schema):
            with validator.resolver.resolving(ref) as resolved:
                for error in validator.descend(instance, resolved):
                    yield error
    else:
        with in_scope(validator.resolver, schema):
            scope, resolved = validator.resolver.resolve(ref)
        validator.resolver.push_scope(scope)

        try:
            for error in validator.descend(instance, resolved):
                yield error
        finally:
            validator.resolver.pop_scope()


@memoize_by_id
def get_validator_type(swagger_spec):
    """Create a custom jsonschema validator for Swagger 2.0 specs.

    :rtype: Its complicated. See jsonschema.validators.create()
    """
    return validators.extend(
        Draft4Validator,
        {
            '$ref': ref_validator,
            'required': functools.partial(required_validator, swagger_spec),
            'enum': functools.partial(enum_validator, swagger_spec),
            'type': functools.partial(type_validator, swagger_spec),
            'format': functools.partial(format_validator, swagger_spec),
            'discriminator': functools.partial(discriminator_validator, swagger_spec),
        },
    )



# -*- coding: utf-8 -*-
import logging

import six

from bravado_core.exception import SwaggerSchemaError

log = logging.getLogger(__name__)


class SecurityRequirement(object):
    """
    Wrapper of security requirement object (http://swagger.io/specification/#securityRequirementObject)

    :param swagger_spec: Spec object
    :type swagger_spec: bravado_core.spec.Spec
    :param security_requirement_spec: security requirement specification in dict form
    """

    def __init__(self, swagger_spec, security_requirement_spec):
        self.swagger_spec = swagger_spec
        self.security_requirement_spec = swagger_spec.deref(security_requirement_spec)
        for security_definition in six.iterkeys(security_requirement_spec):
            if security_definition not in self.swagger_spec.security_definitions:
                raise SwaggerSchemaError(
                    '{security} not defined in {swagger_path}'.format(
                        swagger_path='/securityDefinitions',
                        security=security_definition,
                    ),
                )

    @property
    def security_definitions(self):
        return dict(
            (security_name, self.swagger_spec.security_definitions[security_name])
            for security_name in six.iterkeys(self.security_requirement_spec)
        )

    @property
    def security_scopes(self):
        return dict(
            (security_name, self.security_requirement_spec[security_name])
            for security_name in six.iterkeys(self.security_requirement_spec)
        )

    @property
    def parameters_representation_dict(self):
        return [
            definition.parameter_representation_dict
            for definition in six.itervalues(self.security_definitions)
            if definition.parameter_representation_dict
        ]

    def __iter__(self):
        return six.itervalues(self.security_definitions)



# -*- coding: utf-8 -*-
import sys

import six


class SwaggerError(Exception):
    """Base exception class which all bravado-core specific exceptions
    inherit from.
    """


class SwaggerMappingError(SwaggerError):
    """Raised when an error is encountered during processing of a request or
    a response.
    """


class MatchingResponseNotFound(SwaggerMappingError):
    """Raised when an incoming or outgoing response cannot be matched to a
    documented response in the swagger spec.
    """


class SwaggerValidationError(SwaggerMappingError):
    """Raised when an error is encountered during validating user defined
    format values in a request or a response.
    """


class SwaggerSchemaError(SwaggerError):
    """Raised when an error is encountered during processing of a SwaggerSchema.
    """


class SwaggerSecurityValidationError(SwaggerValidationError):
    """Raised when an error is encountered during processing of
    security related Swagger definitions
    """


def wrap_exception(exception_class):
    """Helper decorator method to modify the raised exception class to
    `exception_class` but keeps the message and trace intact.

    :param exception_class: class to wrap raised exception with
    """
    def generic_exception(method):
        def wrapper(*args, **kwargs):
            try:
                return method(*args, **kwargs)
            except Exception as e:
                six.reraise(
                    exception_class,
                    exception_class(str(e)),
                    sys.exc_info()[2],
                )
        return wrapper
    return generic_exception



# -*- coding: utf-8 -*-
import logging

from six import iteritems

from bravado_core.exception import SwaggerSchemaError
from bravado_core.param import Param
from bravado_core.security_requirement import SecurityRequirement
from bravado_core.util import AliasKeyDict
from bravado_core.util import cached_property
from bravado_core.util import sanitize_name

log = logging.getLogger(__name__)


def _sanitize_operation_id(operation_id, http_method, path_name):
    sanitized_operation_id = sanitize_name(operation_id or '')

    # Handle crazy corner cases where someone explicitly sets operation
    # id a value that gets sanitized down to an empty string or an underscore
    if sanitized_operation_id in {'', '_'}:
        # build based on the http method and request path
        sanitized_operation_id = sanitize_name(http_method + '_' + path_name)

    # Handle super crazy corner case where even ``http_method + '_' + path_name``
    # gets sanitized down to just an underscore (the empty string is just there for good measure)
    if sanitized_operation_id in {'', '_'}:
        # This case is theoretically possible only in case http_method and path_name are
        # sanitized down to empty string. According to the specs valid values of
        # http_method will not allow this case.
        raise ValueError(
            '_sanitize_operation_id produced an empty operation id starting from '
            'operation_id={operation_id}, http_method={http_method} and path_name={path_name}'.format(
                operation_id=operation_id,
                http_method=http_method,
                path_name=path_name,
            ),
        )

    return sanitized_operation_id


class Operation(object):

    def __init__(self, swagger_spec, path_name, http_method, op_spec):
        """Swagger operation defined by a unique (http_method, path_name) pair.

        :type swagger_spec: :class:`Spec`
        :param path_name: path of the operation. e.g. /pet/{petId}
        :param http_method: get/put/post/delete/etc
        :param op_spec: operation specification in dict form
        """
        self.swagger_spec = swagger_spec
        self.path_name = path_name
        self.http_method = http_method
        self.op_spec = swagger_spec.deref(op_spec)

        # (key, value) = (param name, Param)
        self.params = {}

    @cached_property
    def consumes(self):
        """Note that the operation can override the value defined globally
        at #/consumes.

        :return: List of supported mime types consumed by this operation. e.g.
            ["application/x-www-form-urlencoded"]
        :rtype: list of strings, never None
        """
        deref = self.swagger_spec.deref
        result = deref(self.op_spec.get('consumes'))
        if result is None:
            result = deref(self.swagger_spec._internal_spec_dict.get('consumes', []))
        return result

    @cached_property
    def security_specs(self):
        deref = self.swagger_spec.deref
        op_spec = deref(self.op_spec)
        spec_dict = deref(self.swagger_spec._internal_spec_dict)
        if 'security' in op_spec:
            return deref(op_spec['security'])
        else:
            return spec_dict.get('security', [])

    @cached_property
    def security_requirements(self):
        return [
            SecurityRequirement(self.swagger_spec, security_item)
            for security_item in self.security_specs
        ]

    @property
    def acceptable_security_definition_combinations(self):
        return [sorted(security_item.keys()) for security_item in self.security_specs]

    @cached_property
    def security_parameters(self):
        return [
            Param(self.swagger_spec, self, parameter_dict)
            for security_requirement in self.security_requirements
            for parameter_dict in security_requirement.parameters_representation_dict
        ]

    @cached_property
    def produces(self):
        """Note that the operation can override the value defined globally
        at #/produces.

        :return: List of supported mime types produced by this operation. e.g.
            ["application/json"]
        :rtype: list of strings, never None
        """
        deref = self.swagger_spec.deref
        result = deref(self.op_spec.get('produces'))
        if result is None:
            return deref(self.swagger_spec._internal_spec_dict.get('produces', []))
        return result

    @classmethod
    def from_spec(cls, swagger_spec, path_name, http_method, op_spec):
        """
        Creates a :class:`Operation` and builds up its list of :class:`Param` s

        :param swagger_spec: :class:`Spec`
        :param path_name: path of the operation. e.g. /pet/{petId}
        :param http_method: get/put/post/delete/etc
        :param op_spec: operation specification in dict form
        :rtype: :class:`Operation`
        """
        op = cls(swagger_spec, path_name, http_method, op_spec)
        op.params = build_params(op)
        return op

    @cached_property
    def operation_id(self):
        """A friendly name for the operation. The id MUST be unique among all
        operations described in the API. Tools and libraries MAY use the
        operation id to uniquely identify an operation.

        This this field is not required, it will be generated when needed.

        :rtype: str
        """
        deref = self.swagger_spec.deref
        op_id = deref(self.op_spec.get('operationId'))
        return _sanitize_operation_id(op_id, self.http_method, self.path_name)

    def __repr__(self):
        return u"%s(%s)" % (self.__class__.__name__, self.operation_id)


def build_params(op):
    """Builds up the list of this operation's parameters taking into account
    parameters that may be available for this operation's path component.

    :type op: :class:`bravado_core.operation.Operation`

    :returns: dict where (k,v) is (param_name, Param)
    """
    swagger_spec = op.swagger_spec
    deref = swagger_spec.deref
    op_spec = deref(op.op_spec)
    op_params_spec = deref(op_spec.get('parameters', []))
    spec_dict = deref(swagger_spec._internal_spec_dict)
    paths_spec = deref(spec_dict.get('paths', {}))
    path_spec = deref(paths_spec.get(op.path_name))
    path_params_spec = deref(path_spec.get('parameters', []))

    # Order of addition is *important* here. Since op_params are last in the
    # list, they will replace any previously defined path_params with the
    # same name when the final params dict is constructed in the loop below.
    params_spec = path_params_spec + op_params_spec

    params = AliasKeyDict()
    for param_spec in params_spec:
        param = Param(swagger_spec, op, deref(param_spec))
        sanitized_name = sanitize_name(param.name)
        params[sanitized_name] = param
        params.add_alias(param.name, sanitized_name)

    # Security parameters cannot override and been overridden by operation or path objects
    new_params = {}
    new_param_aliases = {}
    for parameter in op.security_parameters:
        param_name = sanitize_name(parameter.name)
        if param_name in params:
            raise SwaggerSchemaError(
                "'{0}' security parameter is overriding a parameter defined in operation or path object".format(
                    parameter.name,
                ),
            )
        else:
            # not directly in params because different security requirements could share parameters
            new_params[param_name] = parameter
            new_param_aliases[parameter.name] = param_name

    params.update(new_params)
    for alias, name in iteritems(new_param_aliases):
        params.add_alias(alias, name)
    return params



# -*- coding: utf-8 -*-
import typing
from mypy_extensions import Arg

try:
    from typing import NoReturn
except ImportError:
    NoReturn = None  # type: ignore


Func = typing.Callable[..., typing.Any]
FuncType = typing.TypeVar('FuncType', bound=Func)
JSONDict = typing.Dict[typing.Text, typing.Any]
MarshalingMethod = typing.Callable[[Arg(typing.Any, 'value')], typing.Any]
UnmarshalingMethod = typing.Callable[[Arg(typing.Any, 'value')], typing.Any]



# -*- coding: utf-8 -*-
import logging

log = logging.getLogger(__name__)


class SecurityDefinition(object):
    """
    Wrapper of security definition object (http://swagger.io/specification/#securityDefinitionsObject)

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type security_definition_spec: security definition specification in dict form
    """

    def __init__(self, swagger_spec, security_definition_spec):
        self.swagger_spec = swagger_spec
        self.security_definition_spec = swagger_spec.deref(security_definition_spec)

    @property
    def location(self):
        # not using 'in' as the name since it is a keyword in python
        return self.security_definition_spec.get('in')

    @property
    def type(self):
        return self.security_definition_spec['type']

    @property
    def name(self):
        return self.security_definition_spec.get('name')

    @property
    def flow(self):
        return self.security_definition_spec.get('flow')

    @property
    def scopes(self):
        return self.security_definition_spec.get('scopes')

    @property
    def authorizationUrl(self):
        return self.security_definition_spec.get('authorizationUrl')

    @property
    def parameter_representation_dict(self):
        if self.type == 'apiKey':
            return {
                'required': False,
                'type': 'string',
                'description': self.security_definition_spec.get('description', ''),
                'name': self.name,
                'in': self.location,
            }



# -*- coding: utf-8 -*-
"""
Support for the 'format' key in the swagger spec as outlined in
https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#dataTypeFormat
"""
import base64
import functools
from collections import namedtuple

import dateutil.parser
import pytz
import six

from bravado_core import schema
from bravado_core.exception import SwaggerMappingError

if six.PY3:
    long = int


def NO_OP(x):
    return x


def to_wire(swagger_spec, primitive_spec, value):
    """Converts a python primitive or object to a reasonable wire
    representation if it has an associated Swagger `format`.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :param primitive_spec: spec for a primitive type as a dict
    :param value: primitive to convert to wire representation
    :type value: int, long, float, boolean, string, unicode, object, etc
    :rtype: int, long, float, boolean, string, unicode, etc
    :raises: SwaggerMappingError when format.to_wire raises an exception
    """
    if value is None or not schema.has_format(swagger_spec, primitive_spec):
        return value
    format_name = schema.get_format(swagger_spec, primitive_spec)
    formatter = swagger_spec.get_format(format_name)

    try:
        return formatter.to_wire(value) if formatter else value
    except Exception as e:
        raise SwaggerMappingError(
            'Error while marshalling value={} to type={}{}.'.format(
                value, primitive_spec['type'],
                '/{}'.format(primitive_spec['format']) if 'format' in primitive_spec else '',
            ),
            e,
        )


def to_python(swagger_spec, primitive_spec, value):
    """Converts a value in wire format to its python representation if
     it has an associated Swagger `format`.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :param primitive_spec: spec for a primitive type as a dict
    :type value: int, long, float, boolean, string, unicode, etc
    :rtype: int, long, float, boolean, string, object, etc
    """
    if value is None or not schema.has_format(swagger_spec, primitive_spec):
        return value
    format_name = schema.get_format(swagger_spec, primitive_spec)
    formatter = swagger_spec.get_format(format_name)
    return formatter.to_python(value) if formatter else value


class SwaggerFormat(
    namedtuple(
        'SwaggerFormat',
        'format to_python to_wire validate description',
    ),
):
    """User-defined format which can be registered with a
    :class:`bravado_core.spec.Spec` to handle marshalling to wire format,
    unmarshalling to a python type, and format specific validation.

    :param format: Name for the user-defined format.
    :param to_python: function to unmarshal a value of this format.
        Eg. lambda val_str: base64.b64decode(val_str)
    :param to_wire: function to marshal a value of this format
        Eg. lambda val_py: base64.b64encode(val_py)
    :param validate: function to validate the correctness of the `wire` value.
        It should raise :class:`bravado_core.exception.SwaggerValidationError`
        if the value does not conform to the format.
    :param description: Short description of the format and conversion logic.
    """


def return_true_wrapper(validate_func):
    """Decorator for the SwaggerFormat.validate function to always return True.

    The contract for `SwaggerFormat.validate` is to raise an exception
    when validation fails. However, the contract for jsonschema's
    validate function is to raise an exception or return True. This wrapper
    bolts-on the `return True` part.

    :param validate_func: SwaggerFormat.validate function
    :return: wrapped callable
    """
    @functools.wraps(validate_func)
    def wrapper(validatable_primitive):
        validate_func(validatable_primitive)
        return True

    return wrapper


BASE64_BYTE_FORMAT = SwaggerFormat(
    format='byte',
    # Note: In Python 3, this requires a bytes-like object as input
    to_wire=lambda b: six.ensure_str(base64.b64encode(b), encoding='ascii'),
    to_python=lambda s: base64.b64decode(
        six.ensure_binary(s, encoding='ascii'),
    ),
    validate=NO_OP,  # jsonschema validates string
    description='Converts [wire]string:byte <=> python bytes',
)

DEFAULT_FORMATS = {
    'byte': SwaggerFormat(
        format='byte',
        to_wire=lambda b: b if isinstance(b, str) else str(b),
        to_python=lambda s: s if isinstance(s, str) else str(s),
        validate=NO_OP,  # jsonschema validates string
        description='Converts [wire]string:byte <=> python byte',
    ),
    'date': SwaggerFormat(
        format='date',
        to_wire=lambda d: d.isoformat(),
        to_python=lambda d: dateutil.parser.parse(d).date(),
        validate=NO_OP,  # jsonschema validates date
        description='Converts [wire]string:date <=> python datetime.date',
    ),
    # Python has no double. float is C's double in CPython
    'double': SwaggerFormat(
        format='double',
        to_wire=lambda d: d if isinstance(d, float) else float(d),
        to_python=lambda d: d if isinstance(d, float) else float(d),
        validate=NO_OP,  # jsonschema validates number
        description='Converts [wire]number:double <=> python float',
    ),
    'date-time': SwaggerFormat(
        format='date-time',
        to_wire=lambda dt: (dt if dt.tzinfo else pytz.utc.localize(dt)).isoformat(),
        to_python=lambda dt: dateutil.parser.parse(dt),
        validate=NO_OP,  # jsonschema validates date-time
        description=(
            'Converts string:date-time <=> python datetime.datetime'
        ),
    ),
    'float': SwaggerFormat(
        format='float',
        to_wire=lambda f: f if isinstance(f, float) else float(f),
        to_python=lambda f: f if isinstance(f, float) else float(f),
        validate=NO_OP,  # jsonschema validates number
        description='Converts [wire]number:float <=> python float',
    ),
    'int32': SwaggerFormat(
        format='int32',
        to_wire=lambda i: i if isinstance(i, int) else int(i),
        to_python=lambda i: i if isinstance(i, int) else int(i),
        validate=NO_OP,  # jsonschema validates integer
        description='Converts [wire]integer:int32 <=> python int',
    ),
    'int64': SwaggerFormat(
        format='int64',
        to_wire=lambda i: i if isinstance(i, long) else long(i),
        to_python=lambda i: i if isinstance(i, long) else long(i),
        validate=NO_OP,  # jsonschema validates integer
        description='Converts [wire]integer:int64 <=> python long',
    ),
}



# -*- coding: utf-8 -*-
"""
The module exposes unmarshaling capabilities.

Unmarshaling is the operation that converts a "JSON" object into its python representation.
The operation should also take care of converting types accordingly to the defined :class:`bravado_core.formatter.SwaggerFormat`s.
"""
import warnings
from functools import partial

import typing
from six import iteritems

from bravado_core import _decorators
from bravado_core import schema
from bravado_core.exception import SwaggerMappingError
from bravado_core.model import MODEL_MARKER
from bravado_core.schema import collapsed_properties
from bravado_core.schema import collapsed_required
from bravado_core.schema import get_type_from_schema
from bravado_core.schema import is_dict_like
from bravado_core.schema import is_list_like
from bravado_core.schema import SWAGGER_PRIMITIVES
from bravado_core.util import memoize_by_id


if getattr(typing, 'TYPE_CHECKING', False):
    from bravado_core._compat_typing import JSONDict
    from bravado_core._compat_typing import NoReturn
    from bravado_core._compat_typing import UnmarshalingMethod
    from bravado_core.model import Model
    from bravado_core.spec import Spec


_NOT_FOUND = object()
_handle_null_value = partial(
    _decorators.handle_null_value,
    is_marshaling_operation=False,
)


def unmarshal_schema_object(swagger_spec, schema_object_spec, value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """
    Unmarshal the value using the given schema object specification.

    Unmarshalling includes:
    - transform the value according to 'format' if available
    - return the value in a form suitable for use. e.g. conversion to a Model
      type.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type schema_object_spec: dict
    :type value: int, float, long, string, unicode, boolean, list, dict, etc

    :return: unmarshalled value
    :rtype: int, float, long, string, unicode, boolean, list, dict, object (in
        the case of a 'format' conversion', or Model type
    """
    unmarshaling_method = _get_unmarshaling_method(swagger_spec=swagger_spec, object_schema=schema_object_spec)
    return unmarshaling_method(value)


def unmarshal_primitive(swagger_spec, primitive_spec, value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Unmarshal a jsonschema primitive type into a python primitive.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type primitive_spec: dict
    :type value: int, long, float, boolean, string, unicode, etc

    :rtype: int, long, float, boolean, string, unicode, or an object
        based on 'format'
    :raises: SwaggerMappingError
    """
    warnings.warn(
        'unmarshal_primitive will be deprecated in the next major release. '
        'Please use the more general entry-point offered in unmarshal_schema_object',
        DeprecationWarning,
    )
    null_decorator = _handle_null_value(swagger_spec=swagger_spec, object_schema=primitive_spec)
    unmarshal_function = _unmarshaling_method_primitive_type(swagger_spec, primitive_spec)

    return null_decorator(unmarshal_function)(value)


def unmarshal_array(swagger_spec, array_spec, array_value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Unmarshal a jsonschema type of 'array' into a python list.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type array_spec: dict
    :type array_value: list
    :rtype: list
    :raises: SwaggerMappingError
    """
    warnings.warn(
        'unmarshal_array will be deprecated in the next major release. '
        'Please use the more general entry-point offered in unmarshal_schema_object',
        DeprecationWarning,
    )
    null_decorator = _handle_null_value(swagger_spec=swagger_spec, object_schema=array_spec)
    unmarshal_function = _unmarshaling_method_array(swagger_spec, array_spec)
    return null_decorator(unmarshal_function)(array_value)


def unmarshal_object(swagger_spec, object_spec, object_value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Unmarshal a jsonschema type of 'object' into a python dict.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type object_spec: dict
    :type object_value: dict
    :rtype: dict
    :raises: SwaggerMappingError
    """
    warnings.warn(
        'unmarshal_object will be deprecated in the next major release. '
        'Please use the more general entry-point offered in unmarshal_schema_object',
        DeprecationWarning,
    )
    null_decorator = _handle_null_value(swagger_spec=swagger_spec, object_schema=object_spec)
    unmarshal_function = _unmarshaling_method_object(swagger_spec, object_spec, use_models=False)
    return null_decorator(unmarshal_function)(object_value)


def unmarshal_model(swagger_spec, model_spec, model_value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Unmarshal a dict into a Model instance.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type model_spec: dict
    :type model_value: dict
    :rtype: Model instance
    :raises: SwaggerMappingError
    """
    warnings.warn(
        'unmarshal_model will be deprecated in the next major release. '
        'Please use the more general entry-point offered in unmarshal_schema_object',
        DeprecationWarning,
    )

    null_decorator = _handle_null_value(swagger_spec=swagger_spec, object_schema=model_spec)
    unmarshal_function = _unmarshaling_method_object(swagger_spec, model_spec, use_models=True)
    return null_decorator(unmarshal_function)(model_value)


@_decorators.wrap_recursive_call_exception
@memoize_by_id
def _get_unmarshaling_method(swagger_spec, object_schema, is_nullable=True):
    # type: (Spec, JSONDict, bool) -> UnmarshalingMethod
    """
    Determine the method needed to unmarshal values of a defined object_schema
    The returned method will accept a single positional parameter that represent the value
    to be unmarshaled.

    :param swagger_spec: Spec object
    :param object_schema: Schema of the object type
    # TODO: remove is_nullable support once https://github.com/Yelp/bravado-core/issues/335 is addressed
    :param is_nullable: Flag set to `True` if the current schema is nullable.
                        The flag will be set to `True` if the schema is not required or `x-nullable`
                        attribute is set to true by the "parent" schema
    """
    object_schema = swagger_spec.deref(object_schema)
    null_decorator = _handle_null_value(
        swagger_spec=swagger_spec,
        object_schema=object_schema,
        is_nullable=is_nullable,
    )
    object_type = get_type_from_schema(swagger_spec, object_schema)

    if object_type == 'array':
        return null_decorator(_unmarshaling_method_array(swagger_spec, object_schema))
    elif object_type == 'file':
        # TODO: Type file is not a valid type. It is present to support parameter unmarshaling (move to bravado_core.param)  # noqa: E501
        return null_decorator(_unmarshaling_method_file(swagger_spec, object_schema))
    elif object_type == 'object':
        return null_decorator(_unmarshaling_method_object(swagger_spec, object_schema))
    elif object_type in SWAGGER_PRIMITIVES:
        return null_decorator(_unmarshaling_method_primitive_type(swagger_spec, object_schema))
    elif object_type is None:
        return _no_op_unmarshaling
    else:
        return partial(
            _unknown_type_unmarshaling,
            object_type,
        )


def _no_op_unmarshaling(value):
    # type: (typing.Any) -> typing.Any
    return value


def _unknown_type_unmarshaling(object_type, value):
    # type: (typing.Union[typing.Type[dict], typing.Type[Model]], typing.Any) -> NoReturn
    raise SwaggerMappingError(
        "Don't know how to unmarshal value {0} with a type of {1}".format(
            value, object_type,
        ),
    )


def _raise_unknown_model(model_name, value):
    # type: (typing.Text, typing.Any) -> NoReturn
    raise SwaggerMappingError('Unknown model {0} when trying to unmarshal {1}'.format(model_name, value))


def _unmarshal_array(unmarshal_array_item_function, value):
    # type: (UnmarshalingMethod, typing.Any) -> typing.Any
    """
    Unmarshal a JSON list to its python representation.

    :param unmarshal_array_item_function: Unmarshaling function for each array item
    :param value: JSON value to unmarshal

    :raises: SwaggerMappingError
    """
    if not is_list_like(value):
        raise SwaggerMappingError('Expected list like type for {0}:{1}'.format(type(value), value))

    return [
        unmarshal_array_item_function(item)
        for item in value
    ]


def _unmarshaling_method_array(swagger_spec, object_schema):
    # type: (Spec, JSONDict) -> UnmarshalingMethod
    """
    Determine the unmarshaling method needed for a schema of a type array.

    The method will be responsible for the identification of the unmarshaling method of the array items.

    :param swagger_spec: Spec object
    :param object_schema: Schema of the object type
    """
    item_schema = swagger_spec.deref(swagger_spec.deref(object_schema).get('items', _NOT_FOUND))
    if item_schema is _NOT_FOUND:
        return _no_op_unmarshaling

    return partial(
        _unmarshal_array,
        _get_unmarshaling_method(swagger_spec=swagger_spec, object_schema=item_schema),
    )


def _unmarshaling_method_file(swagger_spec, object_schema):
    # type: (Spec, JSONDict) -> UnmarshalingMethod
    # TODO: Type file is not a valid type. It is present to support parameter unmarshaling (move to bravado_core.param)  # noqa: E501
    return _no_op_unmarshaling


def _unmarshal_object(
    swagger_spec,  # type: Spec
    model_type,  # type: typing.Union[typing.Type[JSONDict], typing.Type[Model]]
    properties_to_unmarshaling_function,  # type: typing.Dict[typing.Text, UnmarshalingMethod]
    additional_properties_unmarshaling_function,  # type: UnmarshalingMethod
    properties_to_default_value,  # type: JSONDict
    discriminator_property,  # type: typing.Optional[typing.Text]
    possible_discriminated_type_name_to_model,  # type: typing.Dict[typing.Text, Model]
    model_value,  # type: typing.Any
):
    # type: (...) -> typing.Any
    """
    Unmarshal a dict into a Model instance or a dictionary (according to the 'use_models' swagger_spec configuration).

    :param swagger_spec: Spec object
    :param model_type: Type of the return value (:class:`dict` or a subclass of :class:`bravado_core.model.Model`)
    :param properties_to_unmarshaling_function: Mapping between property name and associated unmarshaling method
    :param additional_properties_unmarshaling_function: Unmarshaling function of eventual additional properties
    :param properties_to_default_value: Mapping between property name and the associated unmarshaled default value
    :param discriminator_property: Discriminator property name. It will be `None` if the schema is not a polymorphic schema
    :param possible_discriminated_type_name_to_model: Mapping of the possible dereferenced Model names and Model instances.
    :param model_value: JSON value to unmarshal

    :raises: SwaggerMappingError
    """
    if not is_dict_like(model_value):
        raise SwaggerMappingError(
            "Expected type to be dict for value {0} to unmarshal to a {1}."
            "Was {2} instead.".format(model_value, model_type, type(model_value)),
        )

    if discriminator_property:
        discriminator_value = model_value[discriminator_property]
        discriminated_model = possible_discriminated_type_name_to_model.get(discriminator_value)
        if discriminated_model is not None:
            unmarshal_func = _get_unmarshaling_method(
                swagger_spec=swagger_spec,
                object_schema=discriminated_model._model_spec,
            )
            return unmarshal_func(model_value)

    unmarshaled_value = model_type()
    for property_name, property_value in iteritems(model_value):
        unmarshaling_function = properties_to_unmarshaling_function.get(
            property_name, additional_properties_unmarshaling_function,
        )
        unmarshaled_value[property_name] = unmarshaling_function(property_value)

    if swagger_spec.config['include_missing_properties']:
        for property_name, unmarshaling_function in iteritems(properties_to_unmarshaling_function):
            if property_name not in unmarshaled_value:
                unmarshaled_value[property_name] = properties_to_default_value.get(property_name)

    return unmarshaled_value


def _unmarshaling_method_object(swagger_spec, object_schema, use_models=True):
    # type: (Spec, JSONDict, bool) -> UnmarshalingMethod
    """
    Determine the unmarshaling method needed for a schema of a type object.

    The method will be responsible for the identification of:
     * required, nullable and default value for the all properties
     * unmarshaling methods of all the properties
     * unmarshaling method of the eventual additional properties
     * polymorphic nature of the object (`discriminator` attribute) and list of the associated models

    :param swagger_spec: Spec object
    :param object_schema: Schema of the object type
    # TODO: use_models parameter should be removed once unmarshal_model function is removed
    :param use_models: Flag that enables or disables the usage of Models
    """

    model_type = None  # type: typing.Optional[typing.Type[Model]]
    object_schema = swagger_spec.deref(object_schema)
    if MODEL_MARKER in object_schema:
        model_name = object_schema[MODEL_MARKER]
        model_type = swagger_spec.definitions.get(model_name)
        if use_models and model_type is None:
            return partial(
                _raise_unknown_model,
                model_name,
            )
        if not use_models:
            model_type = None

    properties = collapsed_properties(object_schema, swagger_spec)
    required_properties = collapsed_required(object_schema, swagger_spec)
    properties_to_unmarshaling_function = {
        prop_name: _get_unmarshaling_method(
            swagger_spec=swagger_spec,
            object_schema=prop_schema,
            is_nullable=prop_schema.get('x-nullable', False) or prop_name not in required_properties,
        )
        for prop_name, prop_schema in iteritems(properties)
    }

    additional_properties_unmarshaling_function = _no_op_unmarshaling
    if object_schema.get('additionalProperties') is not False:
        additional_properties_schema = object_schema.get('additionalProperties', {})
        if additional_properties_schema not in ({}, True):
            additional_properties_unmarshaling_function = _get_unmarshaling_method(
                swagger_spec=swagger_spec,
                object_schema=additional_properties_schema,
                is_nullable=False,
            )

    properties_to_default_value = {
        prop_name: unmarshal_schema_object(
            swagger_spec=swagger_spec,
            schema_object_spec=prop_schema,
            value=schema.get_default(swagger_spec, prop_schema),
        )
        for prop_name, prop_schema in iteritems(properties)
        if schema.has_default(swagger_spec, prop_schema)
    }

    discriminator_property = object_schema.get('discriminator')
    possible_discriminated_type_name_to_model = {}
    if model_type and object_schema.get('discriminator'):
        possible_discriminated_type_name_to_model.update({
            k: v
            for k, v in iteritems(swagger_spec.definitions)
            if model_type and model_type.__name__ in v._inherits_from
        })

    return partial(
        _unmarshal_object,
        swagger_spec,
        model_type if model_type and swagger_spec.config['use_models'] else dict,
        properties_to_unmarshaling_function,
        additional_properties_unmarshaling_function,
        properties_to_default_value,
        discriminator_property,
        possible_discriminated_type_name_to_model,
    )


def _unmarshaling_method_primitive_type(swagger_spec, object_schema):
    # type: (Spec, JSONDict) -> UnmarshalingMethod
    """
    Determine the unmarshaling method needed for a schema of a primitive type.

    The method will be responsible for the identification of the eventual :class:`bravado_core.formatter.SwaggerFormat`
    transformation to apply.

    :param swagger_spec: Spec object
    :param object_schema: Schema of the primitive type
    """
    format_name = schema.get_format(swagger_spec, object_schema)
    swagger_format = swagger_spec.get_format(format_name) if format_name is not None else None
    if swagger_format is not None:
        return swagger_format.to_python
    else:
        return _no_op_unmarshaling



# -*- coding: utf-8 -*-
import json
import logging
import os.path
import warnings

import yaml
from jsonref import JsonRef
from jsonschema import FormatChecker
from jsonschema.validators import RefResolver
from six import iteritems
from six import iterkeys
from six.moves import range
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import urlunparse
from six.moves.urllib.request import url2pathname
from swagger_spec_validator import validator20
from swagger_spec_validator.ref_validators import in_scope

from bravado_core import formatter
from bravado_core.exception import SwaggerSchemaError
from bravado_core.exception import SwaggerValidationError
from bravado_core.formatter import return_true_wrapper
from bravado_core.model import model_discovery
from bravado_core.resource import build_resources
from bravado_core.schema import is_dict_like
from bravado_core.schema import is_list_like
from bravado_core.schema import is_ref
from bravado_core.security_definition import SecurityDefinition
from bravado_core.spec_flattening import flattened_spec
from bravado_core.util import cached_property
from bravado_core.util import memoize_by_id
from bravado_core.util import strip_xscope


log = logging.getLogger(__name__)


CONFIG_DEFAULTS = {
    # On the client side, validate incoming responses
    # On the server side, validate outgoing responses
    'validate_responses': True,

    # On the client side, validate outgoing requests
    # On the server side, validate incoming requests
    'validate_requests': True,

    # Use swagger_spec_validator to validate the swagger spec
    'validate_swagger_spec': True,

    # Use Python classes (models) instead of dicts for #/definitions/{models}
    # On the client side, this applies to incoming responses.
    # On the server side, this applies to incoming requests.
    #
    # NOTE: outgoing requests on the client side and outgoing responses on the
    #       server side can use either models or dicts.
    'use_models': True,

    # List of user-defined formats of type
    # :class:`bravado_core.formatter.SwaggerFormat`. These formats are in
    # addition to the formats already supported by the Swagger 2.0
    # Specification.
    'formats': [],

    # Fill with None all the missing properties during object unmarshalling
    'include_missing_properties': True,

    # What to do when a type is missing
    # If True, set the type to object and validate
    # If False, do no validation
    'default_type_to_object': False,

    # Completely dereference $refs to maximize marshaling and unmarshalling performances.
    # NOTE: this depends on validate_swagger_spec
    'internally_dereference_refs': False,

    # What value to assume for basePath if it is missing from the spec (this
    # config option is ignored if basePath is present in the spec)
    # If True, use the 'path' element of the URL the spec was retrieved from
    # If False, set basePath to '/' (conforms to Swagger 2.0 specification)
    'use_spec_url_for_base_path': False,

    # If False, use str() function for 'byte' format
    # If True, encode/decode base64 data for 'byte' format
    'use_base64_for_byte_format': False,
}


class Spec(object):
    """Represents a Swagger Specification for a service.

    :param spec_dict: Swagger API specification in json-like dict form
    :param origin_url: URL from which the spec was retrieved.
    :param http_client: Used to retrieve the spec via http/https.
    :type http_client: :class:`bravado.http_client.HTTPClient`
    :param config: Configuration dict. See CONFIG_DEFAULTS.
    """

    def __init__(
        self, spec_dict, origin_url=None, http_client=None,
        config=None,
    ):
        self.spec_dict = spec_dict
        self.origin_url = origin_url
        self.http_client = http_client
        self.api_url = None
        self.config = dict(CONFIG_DEFAULTS, **(config or {}))

        # (key, value) = (simple format def name, Model type)
        # (key, value) = (#/ format def ref, Model type)
        self.definitions = {}

        # (key, value) = (simple resource name, Resource)
        # (key, value) = (#/ format resource ref, Resource)
        self.resources = None

        # (key, value) = (simple ref name, param_spec in dict form)
        # (key, value) = (#/ format ref name, param_spec in dict form)
        self.params = None

        # Built on-demand - see get_op_for_request(..)
        self._request_to_op_map = None

        # (key, value) = (format name, SwaggerFormat)
        self.user_defined_formats = {}
        self.format_checker = FormatChecker()

        self.resolver = RefResolver(
            base_uri=origin_url or '',
            referrer=self.spec_dict,
            handlers=self.get_ref_handlers(),
        )

        # spec dict used to build resources, in case internally_dereference_refs config is enabled
        # it will be overridden by the dereferenced specs (by build method). More context in PR#263
        self._internal_spec_dict = spec_dict

    @cached_property
    def client_spec_dict(self):
        """Return a copy of spec_dict with x-scope metadata removed so that it
        is suitable for consumption by Swagger clients.

        You may be asking, "Why is there a difference between the Swagger spec
        a client sees and the one used internally?". Well, as part of the
        ingestion process, x-scope metadata is added to spec_dict so that
        $refs can be de-reffed successfully during request/response validation
        and marshalling. This metadata is specific to the context of the
        server and contains files and paths that are not relevant to the
        client. This is required so the client does not re-use (and in turn,
        re-creates) the invalid x-scope metadata created by the server.

        For example, a section of spec_dict that contains a ref would change
        as follows.

        Before:

          'MON': {
            '$ref': '#/definitions/DayHours',
            'x-scope': [
                'file:///happyhour/api_docs/swagger.json',
                'file:///happyhour/api_docs/swagger.json#/definitions/WeekHours'
            ]
          }

        After:

          'MON': {
            '$ref': '#/definitions/DayHours'
          }

        """
        return strip_xscope(self.spec_dict)

    @classmethod
    def from_dict(cls, spec_dict, origin_url=None, http_client=None, config=None):
        """Build a :class:`Spec` from Swagger API Specification

        :param spec_dict: swagger spec in json-like dict form.
        :param origin_url: the url used to retrieve the spec, if any
        :type  origin_url: str
        :param http_client: http client used to download remote $refs
        :param config: Configuration dict. See CONFIG_DEFAULTS.
        """
        spec = cls(spec_dict, origin_url, http_client, config)
        spec.build()
        return spec

    def _validate_spec(self):
        if self.config['validate_swagger_spec']:
            self.resolver = validator20.validate_spec(
                spec_dict=self.spec_dict,
                spec_url=self.origin_url or '',
                http_handlers=self.get_ref_handlers(),
            )

    def build(self):
        self._validate_spec()

        model_discovery(self)

        if self.config['internally_dereference_refs']:
            # Avoid to evaluate is_ref every time, no references are possible at this time
            self.deref = lambda ref_dict: ref_dict
            self._internal_spec_dict = self.deref_flattened_spec

        for user_defined_format in self.config['formats']:
            self.register_format(user_defined_format)

        self.resources = build_resources(self)

        self.api_url = build_api_serving_url(
            spec_dict=self.spec_dict,
            origin_url=self.origin_url,
            use_spec_url_for_base_path=self.config['use_spec_url_for_base_path'],
        )

    def get_ref_handlers(self):
        """Get mapping from URI schemes to handlers that takes a URI.

        The handlers (callables) are used by the RefResolver to retrieve
        remote specification $refs.

        :returns: dict like {'http': callable, 'https': callable}
        :rtype: dict
        """
        return build_http_handlers(self.http_client)

    def _force_deref(self, ref_dict):
        """Dereference ref_dict (if it is indeed a ref) and return what the
        ref points to.

        :param ref_dict:  {'$ref': '#/blah/blah'}
        :return: dereferenced value of ref_dict
        :rtype: scalar, list, dict
        """
        if ref_dict is None or not is_ref(ref_dict):
            return ref_dict

        # Restore attached resolution scope before resolving since the
        # resolver doesn't have a traversal history (accumulated scope_stack)
        # when asked to resolve.
        with in_scope(self.resolver, ref_dict):
            _, target = self.resolver.resolve(ref_dict['$ref'])
            return target

    # NOTE: deref gets overridden, if internally_dereference_refs is enabled, after calling build
    deref = _force_deref

    def get_op_for_request(self, http_method, path_pattern):
        """Return the Swagger operation for the passed in request http method
        and path pattern. Makes it really easy for server-side implementations
        to map incoming requests to the Swagger spec.

        :param http_method: http method of the request
        :param path_pattern: request path pattern. e.g. /foo/{bar}/baz/{id}

        :returns: the matching operation or None if a match couldn't be found
        :rtype: :class:`bravado_core.operation.Operation`
        """
        if self._request_to_op_map is None:
            # lazy initialization
            self._request_to_op_map = {}
            base_path = self.spec_dict.get('basePath', '').rstrip('/')
            for resource in self.resources.values():
                for op in resource.operations.values():
                    full_path = base_path + op.path_name
                    key = (op.http_method, full_path)
                    self._request_to_op_map[key] = op

        key = (http_method.lower(), path_pattern)
        return self._request_to_op_map.get(key)

    def register_format(self, user_defined_format):
        """Registers a user-defined format to be used with this spec.

        :type user_defined_format:
            :class:`bravado_core.formatter.SwaggerFormat`
        """
        name = user_defined_format.format
        self.user_defined_formats[name] = user_defined_format
        validate = return_true_wrapper(user_defined_format.validate)
        self.format_checker.checks(name, raises=(SwaggerValidationError,))(validate)

    def get_format(self, name):
        """
        :param name: Name of the format to retrieve
        :rtype: :class:`bravado_core.formatter.SwaggerFormat`
        """
        user_defined_format = self.user_defined_formats.get(name)
        if user_defined_format is None:
            user_defined_format = formatter.DEFAULT_FORMATS.get(name)
            if name == 'byte' and self.config['use_base64_for_byte_format']:
                user_defined_format = formatter.BASE64_BYTE_FORMAT

        if user_defined_format is None:
            warnings.warn(
                message='{0} format is not registered with bravado-core!'.format(name),
                category=Warning,
            )
        return user_defined_format

    @cached_property
    def security_definitions(self):
        security_defs = {}
        for security_name, security_def in iteritems(self.spec_dict.get('securityDefinitions', {})):
            security_defs[security_name] = SecurityDefinition(self, security_def)
        return security_defs

    @cached_property
    def flattened_spec(self):
        """
        Representation of the current swagger specs that could be written to a single file.
        :rtype: dict
        """

        if not self.config['validate_swagger_spec']:
            log.warning(
                'Flattening unvalidated specs could produce invalid specs. '
                'Use it at your risk or enable `validate_swagger_specs`',
            )

        return strip_xscope(
            spec_dict=flattened_spec(swagger_spec=self),
        )

    @cached_property
    def deref_flattened_spec(self):
        deref_spec_dict = JsonRef.replace_refs(self.flattened_spec)

        @memoize_by_id
        def descend(obj):
            # Inline modification of obj
            # This method is needed because JsonRef could produce performance penalties in accessing
            # the proxied attributes
            if isinstance(obj, JsonRef):
                # Extract the proxied value
                # http://jsonref.readthedocs.io/en/latest/#jsonref.JsonRef.__subject__
                return obj.__subject__
            if is_dict_like(obj):
                for key in list(iterkeys(obj)):
                    obj[key] = descend(obj=obj[key])
            elif is_list_like(obj):
                # obj is list like object provided from flattened_spec specs.
                # This guarantees that it cannot be a tuple instance and
                # inline object modification are allowed
                for index in range(len(obj)):
                    obj[index] = descend(obj=obj[index])
            return obj

        try:
            return descend(obj=deref_spec_dict)
        finally:
            # Make sure that all memory allocated, for caching, could be released
            descend.cache.clear()


def is_yaml(url, content_type=None):
    yaml_content_types = {'application/yaml', 'application/x-yaml', 'text/yaml'}

    yaml_file_extensions = {'.yaml', '.yml'}

    if content_type in yaml_content_types:
        return True

    _, ext = os.path.splitext(url)
    if ext.lower() in yaml_file_extensions:
        return True

    return False


def build_http_handlers(http_client):
    """Create a mapping of uri schemes to callables that take a uri. The
    callable is used by jsonschema's RefResolver to download remote $refs.

    :param http_client: http_client with a request() method

    :returns: dict like {'http': callable, 'https': callable}
    """
    def download(uri):
        log.debug('Downloading %s', uri)
        request_params = {
            'method': 'GET',
            'url': uri,
        }
        response = http_client.request(request_params).result()
        content_type = response.headers.get('content-type', '').lower()
        if is_yaml(uri, content_type):
            return yaml.safe_load(response.content)
        else:
            return response.json()

    def read_file(uri):
        with open(url2pathname(urlparse(uri).path), mode='rb') as fp:
            if is_yaml(uri):
                return yaml.safe_load(fp)
            else:
                return json.loads(fp.read().decode("utf-8"))

    return {
        'http': download,
        'https': download,
        # jsonschema ordinarily handles file:// requests, but it assumes that
        # all files are json formatted. We override it here so that we can
        # load yaml files when necessary.
        'file': read_file,
    }


def build_api_serving_url(
    spec_dict, origin_url=None, preferred_scheme=None, use_spec_url_for_base_path=False,
):
    """The URL used to service API requests does not necessarily have to be the
    same URL that was used to retrieve the API spec_dict.

    The existence of three fields in the root of the specification govern
    the value of the api_serving_url:

    - host string
        The host (name or ip) serving the API. This MUST be the host only and
        does not include the scheme nor sub-paths. It MAY include a port.
        If the host is not included, the host serving the documentation is to
        be used (including the port). The host does not support path templating.

    - basePath string
        The base path on which the API is served, which is relative to the
        host. If it is not included, the API is served directly under the host.
        The value MUST start with a leading slash (/). The basePath does not
        support path templating.

    - schemes [string]
        The transfer protocol of the API. Values MUST be from the list:
        "http", "https", "ws", "wss". If the schemes is not included,
        the default scheme to be used is the one used to access the
        specification.

    See https://github.com/swagger-api/swagger-spec_dict/blob/master/versions/2.0.md#swagger-object-   # noqa

    :param spec_dict: the Swagger spec in json-like dict form
    :param origin_url: the URL from which the spec was retrieved, if any. This
        is only used in Swagger clients.
    :param use_spec_url_for_base_path: only effective when 'basePath' is missing
        from `spec_dict`. When True, 'basePath' will be set to the path portion
        of `origin_url`. When False, 'basePath' will be set to '/'.
    :param preferred_scheme: preferred scheme to use if more than one scheme is
        supported by the API.
    :return: base url which services api requests
    :raises: SwaggerSchemaError
    """
    origin_url = origin_url or 'http://localhost/'
    origin = urlparse(origin_url)

    def pick_a_scheme(schemes):
        if not schemes:
            return origin.scheme

        if preferred_scheme:
            if preferred_scheme in schemes:
                return preferred_scheme
            raise SwaggerSchemaError(
                "Preferred scheme {0} not supported by API. Available schemes "
                "include {1}".format(preferred_scheme, schemes),
            )

        if origin.scheme in schemes:
            return origin.scheme

        return schemes[0]

    netloc = spec_dict.get('host', origin.netloc)
    base_path = '/'
    if use_spec_url_for_base_path:
        base_path = origin.path

    path = spec_dict.get('basePath', base_path)
    scheme = pick_a_scheme(spec_dict.get('schemes'))
    return urlunparse((scheme, netloc, path, None, None, None))



# -*- coding: utf-8 -*-
import copy
import functools
import os.path
import re
import warnings
from collections import defaultdict

from six import iteritems
from six import iterkeys
from six import itervalues
from six.moves.urllib.parse import ParseResult
from six.moves.urllib.parse import urldefrag
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import urlunparse
from six.moves.urllib.request import pathname2url
from six.moves.urllib.request import url2pathname
from swagger_spec_validator.ref_validators import in_scope

from bravado_core.model import model_discovery
from bravado_core.model import MODEL_MARKER
from bravado_core.schema import is_dict_like
from bravado_core.schema import is_list_like
from bravado_core.schema import is_ref
from bravado_core.util import cached_property
from bravado_core.util import determine_object_type
from bravado_core.util import ObjectType


MARSHAL_REPLACEMENT_PATTERNS = {
    '/': '..',  # / is converted to .. (ie. api_docs/swagger.json -> api_docs..swagger.json)
    '#': '|',  # # is converted to | (ie. swagger.json#definitions -> swagger.json|definitions)
}


def _marshal_uri(target_uri, origin_uri):
    """
    Translate the URL string representation into a new string which could be used as JSON keys.
    This method is needed because many JSON parsers and reference resolvers are using '/' as
    indicator of object nesting.

    To workaround this limitation we can re-write the url representation in a way that the parsers
    will accept it, for example "#/definitions/data_type" could become "|..definitions..data_type"

    Example: Assume that you have the following JSON document
        {
            "definitions": {
                "a/possible/def": {
                    "type": "object"
                },
                "a": {
                    "possible": {
                        "def": {
                            "type": "string"
                        }
                    }
                },
                "def": {
                    "$ref": "#/definitions/a/possible/def"
                }
            }
        }

    Assuming that the JSON parser is not raising exception the dereferenced value of
    "#/definitions/def" could be {"type": "object"} or {"type": "string"} which is
    an undetermined condition which can lead to weird errors.
    Let's assume instead that the JSON parser will raise an exception in this case
    the JSON object will not be usable.

    To prevent this conditions we are removing possible '/' from the JSON keys.

    :param target_uri: URI to marshal
    :type target_uri: ParseResult
    :param origin_uri: URI of the root swagger spec file
    :type origin_uri: ParseResult

    :return: a string representation of the URL which could be used into the JSON keys
    :rtype: str
    """

    marshalled_target = urlunparse(target_uri)

    if marshalled_target and target_uri.scheme == '':  # scheme is empty for relative paths. It should NOT happen!
        target_uri = ParseResult('file', *target_uri[1:])
        marshalled_target = urlunparse(target_uri)

    if not marshalled_target or target_uri.scheme not in {'file', 'http', 'https'}:
        raise ValueError(
            'Invalid target: \'{target_uri}\''.format(target_uri=urlunparse(target_uri)),
        )

    if origin_uri and target_uri.scheme == 'file':
        scheme, netloc, path, params, query, fragment = target_uri

        # Masquerade the absolute file path on the "local" server using
        # relative paths from the root swagger spec file
        spec_dir = os.path.dirname(url2pathname(origin_uri.path))
        scheme = 'lfile'
        path = pathname2url(os.path.relpath(url2pathname(path), spec_dir))
        marshalled_target = urlunparse((scheme, netloc, path, params, query, fragment))

    for src, dst in iteritems(MARSHAL_REPLACEMENT_PATTERNS):
        marshalled_target = marshalled_target.replace(src, dst)
    return marshalled_target


class _SpecFlattener(object):
    def __init__(self, swagger_spec, marshal_uri_function):
        self.swagger_spec = swagger_spec
        self.spec_url = self.swagger_spec.origin_url

        if self.spec_url is None:
            warnings.warn(
                message='It is recommended to set origin_url to your spec before flattering it. '
                        'Doing so internal paths will be hidden, reducing the amount of exposed information.',
                category=Warning,
            )

        self.known_mappings = {
            object_type.get_root_holder(): {}
            for object_type in ObjectType
            if object_type.get_root_holder()
        }
        self.marshal_uri_function = marshal_uri_function

    @cached_property
    def marshal_uri(self):
        return functools.partial(
            self.marshal_uri_function,
            origin_uri=urlparse(self.spec_url) if self.spec_url else None,
        )

    @cached_property
    def spec_resolver(self):
        return self.swagger_spec.resolver

    @cached_property
    def resolve(self):
        return self.swagger_spec.resolver.resolve

    @cached_property
    def default_type_to_object(self):
        return self.swagger_spec.config['default_type_to_object']

    def descend(self, value):
        if is_ref(value):
            # Update spec_resolver scope to be able to dereference relative specs from a not root file
            with in_scope(self.spec_resolver, value):
                uri, deref_value = self.resolve(value['$ref'])
                object_type = determine_object_type(
                    object_dict=deref_value,
                    default_type_to_object=self.default_type_to_object,
                )

                known_mapping_key = object_type.get_root_holder()
                if known_mapping_key is None:
                    return self.descend(value=deref_value)
                else:
                    uri = urlparse(uri)
                    if uri not in self.known_mappings.get(known_mapping_key, {}):
                        # The placeholder is present to interrupt the recursion
                        # during the recursive traverse of the data model (``descend``)
                        self.known_mappings[known_mapping_key][uri] = None

                        self.known_mappings[known_mapping_key][uri] = self.descend(value=deref_value)

                    return {'$ref': '#/{}/{}'.format(known_mapping_key, self.marshal_uri(uri))}

        elif is_dict_like(value):
            return {
                key: self.descend(value=subval)
                for key, subval in iteritems(value)
            }

        elif is_list_like(value):
            return [
                self.descend(value=subval)
                for index, subval in enumerate(value)
            ]

        else:
            return value

    def warn_if_uri_clash_on_same_marshaled_representation(self, uri_schema_mappings):
        """
        Verifies that all the uris present on the definitions are represented by a different marshaled uri.
        If is not the case a warning will filed.

        In case of presence of warning please keep us informed about the issue, in the meantime you can
        workaround this calling directly ``flattened_spec(spec, marshal_uri_function)`` passing your
        marshalling function.
        """
        # Check that URIs are NOT clashing to same marshaled representation
        marshaled_uri_mapping = defaultdict(set)
        for uri in iterkeys(uri_schema_mappings):
            marshaled_uri_mapping[self.marshal_uri(uri)].add(uri)

        if len(marshaled_uri_mapping) != len(uri_schema_mappings):
            # At least two uris clashed to the same marshaled representation
            for marshaled_uri, uris in iteritems(marshaled_uri_mapping):
                if len(uris) > 1:
                    warnings.warn(
                        message='{s_uris} clashed to {marshaled}'.format(
                            s_uris=', '.join(sorted(urlunparse(uri) for uri in uris)),
                            marshaled=marshaled_uri,
                        ),
                        category=Warning,
                    )

    def rename_definition_references(self, flattened_spec_dict):
        """
        Rename definition references to more "human" names if possible.

        The used approach is to use model-name as definition key, if this does not conflict
        with an already existing key.

        :param flattened_spec_dict: swagger spec dict (pre-flattened)
        :return: swagger spec dict equivalent to flattened_spec_dict with more human references
        :rtype: dict
        """
        def _rename_references_descend(value):
            if is_ref(value):
                return {
                    '$ref': reference_renaming_mapping.get(value['$ref'], value['$ref']),
                }
            elif is_dict_like(value):
                return {
                    key: _rename_references_descend(value=subval)
                    for key, subval in iteritems(value)
                }

            elif is_list_like(value):
                return [
                    _rename_references_descend(value=subval)
                    for index, subval in enumerate(value)
                ]

            else:
                return value

        definition_key_to_model_name_mapping = {
            k: v[MODEL_MARKER]
            for k, v in iteritems(flattened_spec_dict.get('definitions', {}))
            if is_dict_like(v) and MODEL_MARKER in v
        }

        original_definition_keys = set(iterkeys(flattened_spec_dict.get('definitions', {})))
        new_definition_keys = set(itervalues(definition_key_to_model_name_mapping))

        # Ensure that the new definition keys are not overlapping with already existing ones
        # if this happens the new definition key needs be kept untouched
        reference_renaming_mapping = {
            # old-reference -> new-reference
            '#/definitions/{}'.format(k): '#/definitions/{}'.format(v)
            for k, v in iteritems(definition_key_to_model_name_mapping)
            if v in new_definition_keys and v not in original_definition_keys
        }

        for old_reference, new_reference in iteritems(reference_renaming_mapping):
            new_ref = new_reference.replace('#/definitions/', '')
            old_ref = old_reference.replace('#/definitions/', '')
            flattened_spec_dict['definitions'][new_ref] = flattened_spec_dict['definitions'][old_ref]
            del flattened_spec_dict['definitions'][old_ref]

        return _rename_references_descend(flattened_spec_dict)

    def replace_inline_models_with_refs(self, flattened_spec_dict):
        """
        Rename definition references to more "human" names if possible.

        The used approach is to use model-name as definition key, if this does not conflict
        with an already existing key.

        :param flattened_spec_dict: swagger spec dict (pre-flattened)
        :return: swagger spec dict equivalent to flattened_spec_dict with more human references
        :rtype: dict
        """
        def _set_references_to_models_descend(value, json_ref):
            if is_dict_like(value):
                if (
                    MODEL_MARKER in value and
                    not re.match('^#/definitions/[^/]+$', json_ref) and
                    value == flattened_spec_dict.get('definitions', {}).get(value[MODEL_MARKER])
                ):
                    return {
                        '$ref': '#/definitions/{model_name}'.format(model_name=value[MODEL_MARKER]),
                    }

                else:
                    return {
                        key: _set_references_to_models_descend(value=subval, json_ref='{}/{}'.format(json_ref, key))
                        for key, subval in iteritems(value)
                    }

            elif is_list_like(value):
                return [
                    _set_references_to_models_descend(value=subval, json_ref='{}/{}'.format(json_ref, index))
                    for index, subval in enumerate(value)
                ]

            else:
                return value

        return _set_references_to_models_descend(flattened_spec_dict, '#')

    def model_discovery(self):
        # local imports due to circular dependency
        from bravado_core.spec import Spec

        # Run model-discovery in order to tag the models available in known_mappings['definitions']
        # This is a required step that removes duplications of models due to the presence of models
        # in swagger.json#/definitions and the equivalent models generated by flattening
        model_discovery(
            Spec(
                spec_dict={
                    'definitions': {
                        self.marshal_uri(uri): value
                        for uri, value in iteritems(self.known_mappings['definitions'])
                    },
                },
            ),
        )

    def include_discriminated_models(self):
        """
        This function ensures that discriminated models, present on the original Spec object
        but not directly referenced by the flattened schema (because there is no direct $ref
        attribute) are included in the final flattened specs

        NOTE: The method re-run model_discovery in case additional models have been added to
              the flattened models
        """

        def unflattened_models(flattened_models):
            for m_name, m_type in iteritems(self.swagger_spec.definitions):
                if m_name not in flattened_models:
                    yield m_name, m_type

        def register_unflattened_models():
            """
            :return: True if new models have been added
            """
            initial_number_of_models = len(self.known_mappings['definitions'])
            modified = True

            flattened_models = {
                # schema objects might not have a "type" set so they won't be tagged as models
                definition.get(MODEL_MARKER)
                for definition in itervalues(self.known_mappings['definitions'])
            }

            while modified:
                modified = False
                for model_name, model_type in unflattened_models(flattened_models):
                    if any(
                        parent in flattened_models
                        for parent in model_type._inherits_from
                    ):
                        model_url = urlparse(model_type._json_reference)
                        flattened_models.add(model_name)
                        self.known_mappings['definitions'][model_url] = self.descend(
                            value=model_type._model_spec,
                        )
                        modified = True

            return len(self.known_mappings['definitions']) != initial_number_of_models

        while register_unflattened_models():
            self.model_discovery()

    def include_root_definition(self):
        self.known_mappings['definitions'].update({
            urlparse(v._json_reference): self.descend(value=v._model_spec)
            for v in itervalues(self.swagger_spec.definitions)
            # urldefrag(url)[0] returns the url without the fragment, it is guaranteed to be present
            if urldefrag(v._json_reference)[0] == self.swagger_spec.origin_url
        })

    @cached_property
    def resolved_specs(self):
        # Create internal copy of spec_dict to avoid external dict pollution
        resolved_spec = self.descend(value=copy.deepcopy(self.swagger_spec.spec_dict))

        # Perform model discovery of the newly identified definitions
        self.model_discovery()

        # Ensure that all the root definitions, even if not referenced, are not lost due to flattening.
        self.include_root_definition()

        # Add the identified models that are not available on the know_mappings definitions
        # but that are related, via polymorphism (discriminator), to flattened models
        # This could happen in case discriminated models are not directly referenced by the specs
        # but is fair to assume that they should be on the final artifact due to polymorphism
        self.include_discriminated_models()

        for mapping_key, mappings in iteritems(self.known_mappings):
            self.warn_if_uri_clash_on_same_marshaled_representation(mappings)
            if len(mappings) > 0:
                resolved_spec.update(
                    {
                        mapping_key: {
                            self.marshal_uri(uri): value
                            for uri, value in iteritems(mappings)
                        },
                    },
                )

        resolved_spec = self.rename_definition_references(resolved_spec)
        resolved_spec = self.replace_inline_models_with_refs(resolved_spec)

        return resolved_spec


def flattened_spec(swagger_spec, marshal_uri_function=_marshal_uri):
    """
    Flatten Swagger Specs description into an unique and JSON serializable document.
    The flattening injects in place the referenced [path item objects](https://swagger.io/specification/#pathItemObject)
    while it injects in '#/parameters' the [parameter objects](https://swagger.io/specification/#parameterObject),
    in '#/definitions' the [schema objects](https://swagger.io/specification/#schemaObject) and in
    '#/responses' the [response objects](https://swagger.io/specification/#responseObject).

    Note: the object names in '#/definitions', '#/parameters' and '#/responses' are evaluated by
    ``marshal_uri_function``, the default method takes care of creating unique names for all the used references.
    Since name clashing are still possible take care that a warning could be filed.
    If it happen please report to us the specific warning text and the specs that generated it.
    We can work to improve it and in the mean time you can "plug" a custom marshalling function.

    Note: https://swagger.io/specification/ has been update to track the latest version of the Swagger/OpenAPI specs.
    Please refer to https://github.com/OAI/OpenAPI-Specification/blob/3.0.0/versions/2.0.md#responseObject for the
    most recent Swagger 2.0 specifications.

    WARNING: In the future releases all the parameters except swagger_spec and marshal_uri_function will be removed.
    Please make sure to use only those two parameters.
    Until the deprecation is not effective you can still pass all the parameters but it's strongly discouraged.

    :param swagger_spec: bravado-core Spec object
    :type swagger_spec: bravado_core.spec.Spec
    :param marshal_uri_function: function used to marshal uris in string suitable to be keys in Swagger Specs.
    :type marshal_uri_function: Callable with the same signature of ``_marshal_uri``

    :return: Flattened representation of the Swagger Specs
    :rtype: dict
    """
    return _SpecFlattener(swagger_spec, marshal_uri_function).resolved_specs



# -*- coding: utf-8 -*-
import copy
import inspect
import re

import typing
from enum import Enum
from six import iteritems
from six import iterkeys

from bravado_core._compat import get_function_spec
from bravado_core._compat import wraps
from bravado_core.schema import is_dict_like
from bravado_core.schema import is_list_like

if getattr(typing, 'TYPE_CHECKING', False):
    from bravado_core._compat_typing import FuncType
    CacheKey = typing.Tuple[typing.Tuple[typing.Text, int], ...]


SANITIZE_RULES = [
    (re.compile(regex), replacement)
    for regex, replacement in [
        ('[^A-Za-z0-9_]', '_'),  # valid chars for method names
        ('__+', '_'),  # collapse consecutive _
        ('^[0-9_]+|_+$', ''),  # trim leading/trailing _ and leading digits
    ]
]


FALLBACK_SANITIZE_RULES = [
    (re.compile(regex), replacement)
    for regex, replacement in [
        ('[^A-Za-z0-9_]', '_'),  # valid chars for method names
        ('^[0-9]+', '_'),  # replace leading digits
        ('__+', '_'),  # collapse consecutive _
    ]
]


class cached_property(object):
    """
    A property that is only computed once per instance and then replaces
    itself with an ordinary attribute. Deleting the attribute resets the
    property.

    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
    """

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


class lazy_class_attribute(object):
    """
    An attribute that is computed once per class.

    WARNING: once the attribute has been evaluated is not possible to modify it
    """

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func
        super(lazy_class_attribute, self).__init__()

    def __get__(self, obj, cls):
        value = self.func(cls)
        setattr(cls, self.func.__name__, value)
        return value


class RecursiveCallException(Exception):
    pass


def memoize_by_id(func):
    # type: (FuncType) -> FuncType
    cache = func.cache = {}  # type: ignore  # It's not worth to modify the signature to include handling of cache attribute  # noqa: E501
    key_in_progress_set = set()  # type: typing.Set[CacheKey]
    _CACHE_MISS = object()

    spec = get_function_spec(func)
    default_mapping = dict(zip(reversed(spec.args), reversed(spec.defaults or [])))

    def make_key(*args, **kwargs):
        # type: (typing.Any, typing.Any) -> CacheKey
        """
        Create a cache key starting from *args and **kwargs.
        The cache key is an ordered tuple containing parameter name and related id as in (name, id(value)).
        """
        if args:
            param_name_to_value_mapping = sorted(iteritems(inspect.getcallargs(func, *args, **kwargs)))
        else:
            # Use the default values while determining the parameter name to value to be used for the cache key
            param_name_to_value_mapping = sorted(iteritems(dict(default_mapping, **kwargs)))

        return tuple(
            (param_name, id(param_value))
            for param_name, param_value in param_name_to_value_mapping
        )

    @wraps(func)
    def wrapper(*args, **kwargs):
        # type: (typing.Any, typing.Any) -> typing.Any
        cache_key = make_key(*args, **kwargs)
        cached_value = cache.get(cache_key, _CACHE_MISS)
        if cached_value is _CACHE_MISS:
            if cache_key in key_in_progress_set:
                raise RecursiveCallException()
            key_in_progress_set.add(cache_key)
            cached_value = func(*args, **kwargs)
            key_in_progress_set.remove(cache_key)
            cache[cache_key] = cached_value
        return cached_value
    return wrapper


def sanitize_name(name):
    """Convert a given name so that it is a valid python identifier."""
    if name == '':
        return name

    sanitized_name = name
    for regex, replacement in SANITIZE_RULES:
        sanitized_name = regex.sub(replacement, sanitized_name)

    if sanitized_name == '':  # use fallback rules with more underscores
        sanitized_name = '_' + name  # prepend _ so digits are not stripped
        for regex, replacement in FALLBACK_SANITIZE_RULES:
            sanitized_name = regex.sub(replacement, sanitized_name)

    return sanitized_name


class AliasKeyDict(dict):
    """Dictionary class that allows you to set additional key names for existing keys. Retrieving
    values using these aliased keys works, but when iterating over the dictionary, only the main
    keys are returned."""

    def __init__(self, *args, **kwargs):
        super(AliasKeyDict, self).__init__(*args, **kwargs)
        self.alias_to_key = {}

    def add_alias(self, alias, key):
        if alias != key:
            self.alias_to_key[alias] = key

    def determine_key(self, key):
        if key in self.alias_to_key:  # this will normally be False, optimize for it
            key = self.alias_to_key[key]
        return key

    def get(self, key, *args, **kwargs):
        return super(AliasKeyDict, self).get(self.determine_key(key), *args, **kwargs)

    def pop(self, key, *args, **kwargs):
        return super(AliasKeyDict, self).pop(self.determine_key(key), *args, **kwargs)

    def __getitem__(self, key):
        return super(AliasKeyDict, self).__getitem__(self.determine_key(key))

    def __delitem__(self, key):
        final_key = self.alias_to_key.get(key, key)
        if final_key != key:
            del self.alias_to_key[key]
        return super(AliasKeyDict, self).__delitem__(final_key)

    def __contains__(self, key):
        return super(AliasKeyDict, self).__contains__(self.determine_key(key))

    def copy(self):
        copied_dict = type(self)(self)
        copied_dict.alias_to_key = self.alias_to_key.copy()
        return copied_dict


class ObjectType(Enum):
    """
    Container of object types.
    NOTE: the value of the enum is not supposed to be used outside of this library
    """
    UNKNOWN = None
    SCHEMA = 'definitions'
    PARAMETER = 'parameters'
    RESPONSE = 'responses'
    PATH_ITEM = None

    def get_root_holder(self):
        return self.value


def determine_object_type(object_dict, default_type_to_object=None):
    """
    Use best guess to determine the object type based on the object keys.

    NOTE: it assumes that the base swagger specs are validated and perform type detection for
    the four types of object that could be references in the specs: parameter, path item, response and schema.

    :type object_dict: dict
    :default_type_to_object: Default object type attribute to object if missing (as from bravado_core.spec.Spec config)
    :type default_type_to_object: bool

    :return: determined type of ``object_dict``. The return values is an ObjectType
    :rtype: ObjectType
    """

    if not is_dict_like(object_dict):
        return ObjectType.UNKNOWN

    if 'in' in object_dict and 'name' in object_dict:
        # A parameter object is the only object type that could contain 'in' and 'name' at the same time
        return ObjectType.PARAMETER
    else:
        http_operations = {'get', 'put', 'post', 'delete', 'options', 'head', 'patch'}
        # A path item object MUST have defined at least one http operation and could optionally have 'parameter'
        # attribute. NOTE: patterned fields (``^x-``) are acceptable in path item objects
        object_keys = {key for key in iterkeys(object_dict) if not key.startswith('x-')}
        if object_keys.intersection(http_operations):
            remaining_keys = object_keys.difference(http_operations)
            if not remaining_keys or remaining_keys == {'parameters'}:
                return ObjectType.PATH_ITEM
        else:
            # A response object has:
            #  - mandatory description field
            #  - optional schema, headers and examples field
            #  - no other fields are allowed
            response_allowed_keys = {'description', 'schema', 'headers', 'examples'}

            # If description field is specified and there are no other fields other the allowed response fields
            if 'description' in object_keys and not bool(object_keys - response_allowed_keys):
                return ObjectType.RESPONSE
            else:
                # A schema object has:
                #  - no mandatory parameters
                #  - long list of optional parameters (ie. description, type, items, properties, discriminator, etc.)
                #  - no other fields are allowed
                # NOTE: In case the method is mis-determining the type of a schema object, confusing it with a
                #       response type it will be enough to add, to the object, one key that is not defined
                #       in ``response_allowed_keys``.  (ie. ``additionalProperties: {}``, implicitly defined be specs)
                if default_type_to_object or 'type' in object_dict:
                    return ObjectType.SCHEMA
                else:
                    return ObjectType.UNKNOWN


def strip_xscope(spec_dict):
    """
    :param spec_dict: Swagger spec in dict form. This is treated as read-only.
    :return: deep copy of spec_dict with the x-scope metadata stripped out.
    """
    result = copy.deepcopy(spec_dict)

    def descend(fragment):
        if is_dict_like(fragment):
            fragment.pop('x-scope', None)  # Removes 'x-scope' key if present
            for key in iterkeys(fragment):
                descend(fragment[key])
        elif is_list_like(fragment):
            for element in fragment:
                descend(element)

    descend(result)
    return result



# -*- coding: utf-8 -*-
import warnings
from functools import partial

import typing
from six import iteritems

from bravado_core import _decorators
from bravado_core import schema
from bravado_core.exception import SwaggerMappingError
from bravado_core.model import Model
from bravado_core.model import MODEL_MARKER
from bravado_core.schema import collapsed_properties
from bravado_core.schema import collapsed_required
from bravado_core.schema import get_type_from_schema
from bravado_core.schema import is_dict_like
from bravado_core.schema import is_list_like
from bravado_core.schema import SWAGGER_PRIMITIVES
from bravado_core.util import memoize_by_id


if getattr(typing, 'TYPE_CHECKING', False):
    from bravado_core._compat_typing import JSONDict
    from bravado_core._compat_typing import MarshalingMethod
    from bravado_core._compat_typing import NoReturn
    from bravado_core.spec import Spec
    from bravado_core.formatter import SwaggerFormat


_NOT_FOUND = object()
_handle_null_value = partial(
    _decorators.handle_null_value,
    is_marshaling_operation=True,
)


def marshal_schema_object(swagger_spec, schema_object_spec, value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Marshal the value using the given schema object specification.

    Marshaling includes:
    - transform the value according to 'format' if available
    - return the value in a form suitable for 'on-the-wire' transmission

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type schema_object_spec: dict
    :type value: int, long, string, unicode, boolean, list, dict, Model type

    :return: marshaled value
    :rtype: int, long, string, unicode, boolean, list, dict
    :raises: SwaggerMappingError
    """
    marshaling_method = _get_marshaling_method(swagger_spec=swagger_spec, object_schema=schema_object_spec)
    return marshaling_method(value)


def marshal_primitive(swagger_spec, primitive_spec, value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Marshal a python primitive type into a jsonschema primitive.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type primitive_spec: dict
    :type value: int, long, float, boolean, string, unicode, or an object
        based on 'format'

    :rtype: int, long, float, boolean, string, unicode, etc
    :raises: SwaggerMappingError
    """
    warnings.warn(
        'marshal_primitive will be deprecated in the next major release. '
        'Please use the more general entry-point offered in marshal_schema_object',
        DeprecationWarning,
    )

    null_decorator = _handle_null_value(
        swagger_spec=swagger_spec,
        object_schema=primitive_spec,
    )
    marshal_function = _marshaling_method_primitive_type(swagger_spec=swagger_spec, object_schema=primitive_spec)
    return null_decorator(marshal_function)(value)


def marshal_array(swagger_spec, array_spec, array_value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Marshal a jsonschema type of 'array' into a json-like list.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type array_spec: dict
    :type array_value: list
    :rtype: list
    :raises: SwaggerMappingError
    """
    warnings.warn(
        'marshal_array will be deprecated in the next major release. '
        'Please use the more general entry-point offered in marshal_schema_object',
        DeprecationWarning,
    )
    null_decorator = _handle_null_value(
        swagger_spec=swagger_spec,
        object_schema=array_spec,
    )
    marshal_function = _marshaling_method_array(swagger_spec=swagger_spec, object_schema=array_spec)
    return null_decorator(marshal_function)(array_value)


def marshal_object(swagger_spec, object_spec, object_value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Marshal a python dict to json dict.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type object_spec: dict
    :type object_value: dict

    :rtype: dict
    :raises: SwaggerMappingError
    """
    warnings.warn(
        'marshal_object will be deprecated in the next major release. '
        'Please use the more general entry-point offered in marshal_schema_object',
        DeprecationWarning,
    )
    null_decorator = _handle_null_value(
        swagger_spec=swagger_spec,
        object_schema=object_spec,
    )
    marshal_function = _marshaling_method_object(swagger_spec=swagger_spec, object_schema=object_spec)
    return null_decorator(marshal_function)(object_value)


def marshal_model(swagger_spec, model_spec, model_value):
    # type: (Spec, JSONDict, typing.Any) -> typing.Any
    """Marshal a Model instance into a json-like dict.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type model_spec: dict
    :type model_value: Model instance
    :rtype: dict
    :raises: SwaggerMappingError
    """
    warnings.warn(
        'marshal_model will be deprecated in the next major release. '
        'Please use the more general entry-point offered in marshal_schema_object',
        DeprecationWarning,
    )
    null_decorator = _handle_null_value(
        swagger_spec=swagger_spec,
        object_schema=model_spec,
    )
    marshal_function = _marshaling_method_object(swagger_spec=swagger_spec, object_schema=model_spec)
    return null_decorator(marshal_function)(model_value)


@_decorators.wrap_recursive_call_exception
@memoize_by_id
def _get_marshaling_method(swagger_spec, object_schema, required=False):
    # type: (Spec, JSONDict, bool) -> MarshalingMethod
    """
    Determine the method needed to marshal values of a defined object_schema
    The returned method will accept a single positional parameter that represent the value
    to be marshaled.

    :type swagger_spec: :class:`bravado_core.spec.Spec`
    :type object_schema: dict
    """
    object_schema = swagger_spec.deref(object_schema)
    null_decorator = _handle_null_value(
        swagger_spec=swagger_spec,
        object_schema=object_schema,
        is_nullable=not required,
    )
    object_type = get_type_from_schema(swagger_spec, object_schema)

    if object_type == 'array':
        return null_decorator(_marshaling_method_array(swagger_spec, object_schema))
    elif object_type == 'file':
        return null_decorator(_marshaling_method_file(swagger_spec, object_schema))
    elif object_type == 'object':
        return null_decorator(_marshaling_method_object(swagger_spec, object_schema))
    elif object_type in SWAGGER_PRIMITIVES:
        return null_decorator(_marshaling_method_primitive_type(swagger_spec, object_schema))
    elif object_type is None:
        return _no_op_marshaling
    else:
        return partial(
            _unknown_type_marshaling,
            object_type,
        )


def _no_op_marshaling(value):
    # type: (typing.Any) -> typing.Any
    return value


def _unknown_type_marshaling(object_type, value):
    # type: (typing.Union[typing.Type[dict], typing.Type[Model]], typing.Any) -> NoReturn
    raise SwaggerMappingError(
        'Unknown type {0} for value {1}'.format(
            object_type, value,
        ),
    )


def _raise_unknown_model(model_name, value):
    # type: (typing.Text, typing.Any) -> NoReturn
    raise SwaggerMappingError('Unknown model {0} when trying to marshal {1}'.format(model_name, value))


def _marshal_array(marshal_array_item_function, value):
    # type: (MarshalingMethod, typing.Any) -> typing.Any
    """
    Marshal a python list to its JSON list representation.

    :param marshal_array_item_function: Marshaling function for each array item
    :param value: Python list/tuple to marshal as JSON Array

    :raises: SwaggerMappingError
    """
    if not is_list_like(value):
        raise SwaggerMappingError('Expected list like type for {0}:{1}'.format(type(value), value))

    return [
        marshal_array_item_function(item)
        for item in value
    ]


def _marshaling_method_array(swagger_spec, object_schema):
    # type: (Spec, JSONDict) -> MarshalingMethod
    """
    Determine the marshaling method needed for a schema of a type array.

    The method will be responsible for the identification of the marshaling method of the array items.

    :param swagger_spec: Spec object
    :param object_schema: Schema of the object type
    """
    item_schema = swagger_spec.deref(swagger_spec.deref(object_schema).get('items', _NOT_FOUND))
    if item_schema is _NOT_FOUND:
        return _no_op_marshaling

    return partial(
        _marshal_array,
        _get_marshaling_method(swagger_spec=swagger_spec, object_schema=item_schema),
    )


def _marshaling_method_file(swagger_spec, object_schema):
    # type: (Spec, JSONDict) -> MarshalingMethod
    # TODO: Type file is not a valid type. It is present to support parameter marshaling (move to bravado_core.param)  # noqa: E501
    return _no_op_marshaling


def _marshal_object(
    swagger_spec,  # type: Spec
    properties_to_marshaling_function,  # type: typing.Dict[typing.Text, MarshalingMethod]
    additional_properties_marshaling_function,  # type: MarshalingMethod
    discriminator_property,  # type: typing.Optional[typing.Text]
    possible_discriminated_type_name_to_model,  # type: typing.Dict[typing.Text, Model]
    required_properties,  # type: typing.Set[typing.Text]
    nullable_properties,  # type: typing.Set[typing.Text]
    model_value,  # type: typing.Any
):
    # type: (...) -> typing.Any
    """
    Marshal a dict or Model instance into its JSON Object representation.

    :param swagger_spec: Spec object
    :param properties_to_marshaling_function: Mapping between property name and associated unmarshaling method
    :param additional_properties_marshaling_function: Unmarshaling function of eventual additional properties
    :param discriminator_property: Discriminator property name. It will be `None` if the schema is not a polymorphic schema
    :param possible_discriminated_type_name_to_model: Mapping of the possible dereferenced Model names and Model instances.
    :param required_properties: Set of required properties of the object schema
    :param nullable_properties: Set of nullable properties of the object schema
    :param model_value: Python dictionary or Model to marshal as JSON Object

    :raises: SwaggerMappingError
    """
    if not is_dict_like(model_value) and not isinstance(model_value, Model):
        raise SwaggerMappingError(
            "Expected type to be dict or Model to marshal value '{0}' to a dict. Was {1} instead.".format(
                model_value, type(model_value),
            ),
        )

    if discriminator_property:
        discriminator_value = model_value[discriminator_property]
        discriminated_model = possible_discriminated_type_name_to_model.get(discriminator_value)
        if discriminated_model is not None:
            marshaling_function = _get_marshaling_method(swagger_spec=swagger_spec, object_schema=discriminated_model._model_spec)
            return marshaling_function(model_value)

    marshaled_value = dict()
    for property_name in model_value:
        property_value = model_value[property_name]
        property_marshaling_function = properties_to_marshaling_function.get(
            property_name, additional_properties_marshaling_function,
        )
        if (
            property_value is None
            and property_name not in required_properties
            and property_name not in nullable_properties
            and property_name in properties_to_marshaling_function
        ):
            continue

        marshaled_value[property_name] = property_marshaling_function(property_value)

    return marshaled_value


def _marshaling_method_object(swagger_spec, object_schema):
    # type: (Spec, JSONDict) -> MarshalingMethod
    """
    Determine the marshaling method needed for a schema of a type object.

    The method will be responsible for the identification of:
     * required and nullable value for the all properties
     * marshaling methods of all the properties
     * marshaling method of the eventual additional properties
     * polymorphic nature of the object (`discriminator` attribute) and list of the associated models

    :param swagger_spec: Spec object
    :param object_schema: Schema of the object type
    """

    model_type = None  # type: typing.Optional[typing.Type[Model]]
    object_schema = swagger_spec.deref(object_schema)
    if MODEL_MARKER in object_schema:
        model_name = object_schema[MODEL_MARKER]
        model_type = swagger_spec.definitions.get(model_name)
        if model_type is None:
            return partial(
                _raise_unknown_model,
                model_name,
            )

    properties = collapsed_properties(object_schema, swagger_spec)
    required_properties = collapsed_required(object_schema, swagger_spec)
    properties_to_marshaling_function = {
        prop_name: _get_marshaling_method(
            swagger_spec=swagger_spec,
            object_schema=prop_schema,
            required=prop_name in required_properties,
        )
        for prop_name, prop_schema in iteritems(properties)
    }

    additional_properties_marshaling_function = _no_op_marshaling
    if object_schema.get('additionalProperties') is not False:
        additional_properties_schema = object_schema.get('additionalProperties', {})
        if additional_properties_schema not in ({}, True):
            additional_properties_marshaling_function = _get_marshaling_method(
                swagger_spec=swagger_spec,
                object_schema=additional_properties_schema,
            )

    discriminator_property = object_schema.get('discriminator')
    possible_discriminated_type_name_to_model = {}
    if model_type and object_schema.get('discriminator'):
        possible_discriminated_type_name_to_model.update({
            k: v
            for k, v in iteritems(swagger_spec.definitions)
            if model_type and model_type.__name__ in v._inherits_from
        })

    nullable_properties = {
        prop_name
        for prop_name, prop_schema in iteritems(properties)
        if schema.is_prop_nullable(swagger_spec, prop_schema)
    }

    return partial(
        _marshal_object,
        swagger_spec,
        properties_to_marshaling_function,
        additional_properties_marshaling_function,
        discriminator_property,
        possible_discriminated_type_name_to_model,
        required_properties,
        nullable_properties,
    )


def _marshal_primitive_type(primitive_type, swagger_format, value):
    # type: (typing.Text, SwaggerFormat, typing.Any) -> typing.Any
    """
    Marshal a python representation of a primitive type to its JSON representation.

    :param primitive_type: Primitive type of the schema to marshal
    :param swagger_format: Swagger format to apply during marshaling
    :param value: Python primitive value to marshal
    """
    try:
        return swagger_format.to_wire(value)
    except Exception as e:
        raise SwaggerMappingError(
            'Error while marshalling value={} to type={}/{}.'.format(
                value, primitive_type, swagger_format.format,
            ),
            e,
        )


def _marshaling_method_primitive_type(swagger_spec, object_schema):
    # type: (Spec, JSONDict) -> MarshalingMethod
    """
    Determine the marshaling method needed for a schema of a primitive type.

    The method will be responsible for the identification of the eventual :class:`bravado_core.formatter.SwaggerFormat`
    transformation to apply.

    :param swagger_spec: Spec object
    :param object_schema: Schema of the primitive type
    """
    format_name = schema.get_format(swagger_spec, object_schema)
    swagger_format = swagger_spec.get_format(format_name) if format_name is not None else None
    if swagger_format is not None:
        return partial(
            _marshal_primitive_type,
            get_type_from_schema(swagger_spec, object_schema),
            swagger_format,
        )
    else:
        return _no_op_marshaling
